/*

*/
import scene.*;
import scene.models.*;
import pipeline.*;
import framebuffer.*;

import java.awt.Color;

/**
   Here is the "scene graph" for this example.
   Only the PrismModel node holds any geometry.
   All of the other nodes hold only a matrix.

          LinkedModels
            /    \
           /       \
        link1      link2
           \       /
            \    /
              \/
             mTop
            /|  |\
          /  |  |  \
        /    |  |    \
      m1    m2  m3    m4
        \    |  |    /
          \  |  |  /
            \|  |/
              m0
              |
              |
          PrismModel
*/
public class NestedPrisms_v3
{
   public static void main(String[] args)
   {
      double sqrt3 = Math.sqrt(3.0);

      // Create a Prism object.
      Model m0 = new PrismModel(25, 2.0, Math.PI/4.0);
      m0.setColorRandom();


      // Create four new Prism objects from m0.
      // right
      Model m1 = new Model();
      m1.addSubModel( m0 );
      m1.modelTranslate(2+0.5/sqrt3, 0, 0);
      // left
      Model m2 = new Model();
      m2.addSubModel( m0 );
      m2.modelTranslate(-2-0.5/sqrt3, 0, 0);
      m2.modelRotate(180, 0,0,1);
      // top
      Model m3 = new Model();
      m3.addSubModel( m0 );
      m3.modelRotate(90, 0,0,1);
      m3.modelTranslate(2+0.5/sqrt3, 0, 0);
      // bottom
      Model m4 = new Model();
      m4.addSubModel( m0 );
      m4.modelRotate(-90, 0,0,1);
      m4.modelTranslate(2+0.5/sqrt3, 0, 0);

      // Creat a top level model to hold the four sub models
      Model mTop = new Model();
      mTop.addSubModel(m1);
      mTop.addSubModel(m2);
      mTop.addSubModel(m3);
      mTop.addSubModel(m4);

      // Create two "linked" copies of the top level model.
      Model link1 = new Model();
      link1.addSubModel(mTop);
      Model link2 = new Model();
      link2.addSubModel(mTop);
      link2.modelRotate(180, 0,1,0);
      link2.modelTranslate(2, 0, 0);
      link2.modelRotate(90, 1,0,0);
      Model linkedModels = new Model();
      linkedModels.addSubModel(link1);
      linkedModels.addSubModel(link2);


      Model floor = new PanelModel(-4, 4, -4, 4);
      floor.modelTranslate(0, -4, 0);
      // Push this model away from where the camera is.
      floor.modelTranslate(0, 0, -5);


      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the top level model to the Scene.
      scene.modelList.add(linkedModels);
      scene.modelList.add(floor);


      // Create a Camera object.
      Camera camera = new Camera();

      // Add the Camera to the Scene.
      scene.setCamera(camera);


      // Create a framebuffer to render our scene into.
      int vp_width  = 512;
      int vp_height = 512;
      FrameBuffer fb = new FrameBuffer(vp_width, vp_height);
      // Give the framebuffer a nice background color.
      fb.clearFB(java.awt.Color.lightGray);

      String basename = "PPM_NestedPrism_v3_Frame_";

      // Spin the model 360 degrees arond two axes.
      for (int i = 0; i <= 72; i++)
      {
         // Rotate the top level model 5 degrees.
         linkedModels.model2Identity();
         // Push the model away from where the camera is.
         linkedModels.modelTranslate(0, 0, -6);
         linkedModels.modelRotate( 5*i, 1,0,0);
         linkedModels.modelRotate( 5*i, 0,1,0);
         linkedModels.modelRotate( 5*i, 0,0,1);

         // Render again.
         fb.clearFB(java.awt.Color.lightGray);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format(basename + "%02d.ppm", i));

         // Rotate the right side of each link.
         m1.modelRotate( 5, 1,0,0);  // right side
      }
   }
}
