/*

*/
import scene.*;
import scene.models.*;
import pipeline.*;
import framebuffer.*;

import java.awt.Color;
import java.awt.event.*;

/**
   An extendable robot arm model.
*/
class Arm extends Model
{
   private double shoulderAngle = 0.0;
   private double    elbowAngle = 0.0;
   private double    wristAngle = 0.0;
   private double   fingerAngle = 0.0;

   private double shoulderLength = 1.0;
   private double    elbowLength = 1.0;
   private double    wristLength = 1.0;
   private double   fingerLength = 1.0;

   private Model shoulder; // these are convinient references
   private Model elbow;    // to this model's nested sub models
   private Model wrist;
   private Model finger;

   public Arm()
   {
      super();

      /*
         Be sure to draw a picture of the (simple) tree that this code creates.
      */
      this.addSubModel( new Model() ); // shoulder
      this.getSubModel(0).addLineSegment(new LineSegment(new Vertex(0, 0, 0),
                                                         new Vertex(1, 0, 0)));

      this.getSubModel(0).addSubModel( new Model() ); // elbow
      this.getSubModel(0).getSubModel(0).addLineSegment(new LineSegment(new Vertex(0, 0, 0),
                                                                        new Vertex(1, 0, 0)));

      this.getSubModel(0).getSubModel(0).addSubModel( new Model() ); // wrist
      this.getSubModel(0).getSubModel(0).getSubModel(0).addLineSegment(
                                                          new LineSegment(new Vertex(0, 0, 0),
                                                                          new Vertex(1, 0, 0)));

      this.getSubModel(0).getSubModel(0).getSubModel(0).addSubModel( new Model() ); // finger
      this.getSubModel(0).getSubModel(0).getSubModel(0).getSubModel(0).addLineSegment(
                                                          new LineSegment(new Vertex(0, 0, 0),
                                                                          new Vertex(1, 0, 0)));

      shoulder = this.getSubModel(0);
      elbow    = shoulder.getSubModel(0);
      wrist    = elbow.getSubModel(0);
      finger   = wrist.getSubModel(0);
      updateMatrices();
   }

   public void setShoulder(double length, double angle)
   {
      this.shoulderLength = length;
      this.shoulderAngle = angle;
      updateMatrices();
   }

   public void setElbow(double length, double angle)
   {
      this.elbowLength = length;
      this.elbowAngle = angle;
      updateMatrices();
   }

   public void setWrist(double length, double angle)
   {
      this.wristLength = length;
      this.wristAngle = angle;
      updateMatrices();
   }

   public void setFinger(double length, double angle)
   {
      this.fingerLength = length;
      this.fingerAngle = angle;
      updateMatrices();
   }

   private void updateMatrices()
   {
      shoulder.model2Identity();
      shoulder.modelRotate(shoulderAngle, 0,0,1);
      shoulder.modelScale(shoulderLength, shoulderLength, 1);

      elbow.model2Identity();
      elbow.modelTranslate(1, 0, 0);
      elbow.modelRotate(elbowAngle, 0,0,1);
      elbow.modelScale(elbowLength/shoulderLength, elbowLength/shoulderLength, 1);

      wrist.model2Identity();
      wrist.modelTranslate(1, 0, 0);
      wrist.modelRotate(wristAngle, 0,0,1);
      wrist.modelScale(wristLength/elbowLength, wristLength/elbowLength, 1);

      finger.model2Identity();
      finger.modelTranslate(1, 0, 0);
      finger.modelRotate(fingerAngle, 0,0,1);
      finger.modelScale(fingerLength/wristLength, fingerLength/wristLength, 1);
   }
}


/**
   Draw an interactive robot arm with shoulder, elbow, wrist, and finger joints.
   Make each part of the robot arm extendable.
*/
public class RobotArm4
{
   private static double xTranslation = 0.0;
   private static double yTranslation = 0.0;

   private static double shoulderRotation = 0.0;
   private static double    elbowRotation = 0.0;
   private static double    wristRotation = 0.0;
   private static double   fingerRotation = 0.0;

   private static double shoulderLength = 0.4;
   private static double    elbowLength = 0.3;
   private static double    wristLength = 0.2;
   private static double   fingerLength = 0.1;

   public static void main(String[] args)
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the s/S keys to rotate the arm at the shoulder.");
      System.out.println("Use the e/E keys to rotate the arm at the elbow.");
      System.out.println("Use the w/W keys to rotate the arm at the wrist.");
      System.out.println("Use the f/F keys to rotate the arm at the finger.");
      System.out.println("Use the ^s/^S keys to extend the length of the arm at the shoulder.");
      System.out.println("Use the ^e/^E keys to extend the length of the arm at the elbow.");
      System.out.println("Use the ^w/^W keys to extend the length of the arm at the wrist.");
      System.out.println("Use the ^f/^F keys to extend the length of the arm at the finger.");
      System.out.println("Use the x/X keys to translate the arm along the x-axis.");
      System.out.println("Use the y/Y keys to translate the arm along the y-axis.");

      Arm arm = new Arm();
      arm.setColor( Color.blue );
      // Push the model away from where the camera is.
      arm.modelTranslate(0, 0, -1);

      // Initialize the model matrices for the sub models.
      arm.setShoulder(shoulderLength, shoulderRotation);
      arm.setElbow(elbowLength,       elbowRotation);
      arm.setWrist(wristLength,       wristRotation);
      arm.setFinger(fingerLength    , fingerRotation);


      // Create the Scene object that we shall render
      Scene scene = new Scene();

      // Add the model to the Scene.
      scene.modelList.add(arm);


      // Create a Camera object.
      Camera camera = new Camera();

      // Add the Camera to the Scene.
      scene.setCamera(camera);


      // Create a InteractiveFrame and a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;

      @SuppressWarnings("serial")
      final InteractiveFrame app = new InteractiveFrame("Renderer 6", width, height)
      {
         // Implement part of the KeyListener interface.
         @Override public void keyPressed(KeyEvent e)
         {
          //System.out.println( e );
            if (e.isControlDown() && e.isShiftDown() && e.getKeyCode() == 83) // ^S
            {
               shoulderLength -= 0.02;
            }
            else if (e.isControlDown() && e.getKeyCode() == 83) // ^s
            {
               shoulderLength += 0.02;
            }
            else if (e.isControlDown() && e.isShiftDown() && e.getKeyCode() == 69) // ^E
            {
               elbowLength -= 0.02;
            }
            else if (e.isControlDown() && e.getKeyCode() == 69) // ^e
            {
               elbowLength += 0.02;
            }
            else if (e.isControlDown() && e.isShiftDown() && e.getKeyCode() == 87) // ^W
            {
               wristLength -= 0.02;
            }
            else if (e.isControlDown() && e.getKeyCode() == 87)  // ^w
            {
               wristLength += 0.02;
            }
            else if (e.isControlDown() && e.isShiftDown() && e.getKeyCode() == 70) // ^F
            {
               fingerLength -= 0.02;
            }
            else if (e.isControlDown() && e.getKeyCode() == 70) // ^f
            {
               fingerLength += 0.02;
            }

            // Update the model matrices for the sub models.
            arm.setShoulder(shoulderLength, shoulderRotation);
            arm.setElbow(elbowLength, elbowRotation);
            arm.setWrist(wristLength, wristRotation);
            arm.setFinger(fingerLength, fingerRotation);

            // Render again.
            this.getFrameBuffer().clearFB(Color.green);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }

         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
          //System.out.println( e );
            char c = e.getKeyChar();
            if ('d' == c)
            {
             //Pipeline.debug = ! Pipeline.debug;
               Pipeline.debug_visible = ! Pipeline.debug_visible;
             //Rasterize.debug = ! Rasterize.debug;
             //RasterizeAntialias.debug = ! RasterizeAntialias.debug;
            }
            else if ('a' == c)
            {
               Pipeline.doAntialiasing = ! Pipeline.doAntialiasing;
            }
            else if ('x' == c)
            {
               xTranslation += 0.02;
            }
            else if ('X' == c)
            {
               xTranslation -= 0.02;
            }
            else if ('y' == c)
            {
               yTranslation += 0.02;
            }
            else if ('Y' == c)
            {
               yTranslation -= 0.02;
            }
            else if ('s' == c)
            {
               shoulderRotation += 2.0;
            }
            else if ('S' == c)
            {
               shoulderRotation -= 2.0;
            }
            else if ('e' == c)
            {
               elbowRotation += 2.0;
            }
            else if ('E' == c)
            {
               elbowRotation -= 2.0;
            }
            else if ('w' == c)
            {
               wristRotation += 2.0;
            }
            else if ('W' == c)
            {
               wristRotation -= 2.0;
            }
            else if ('f' == c)
            {
               fingerRotation += 2.0;
            }
            else if ('F' == c)
            {
               fingerRotation -= 2.0;
            }

            arm.model2Identity();
            arm.modelTranslate(0, 0, -1);
            arm.modelTranslate(xTranslation, yTranslation, 0);
            // Update the model matrices for the sub models.
            arm.setShoulder(shoulderLength, shoulderRotation);
            arm.setElbow(elbowLength, elbowRotation);
            arm.setWrist(wristLength, wristRotation);
            arm.setFinger(fingerLength, fingerRotation);

            // Render again.
            this.getFrameBuffer().clearFB(Color.green);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
          //System.out.println( e );
            java.awt.Container c = (java.awt.Container)e.getSource();
            java.awt.Insets insets = c.getInsets();
            int w = c.getWidth()  - insets.left - insets.right;
            int h = c.getHeight() - insets.top  - insets.bottom;

            this.createFrameBuffer(w, h);
            this.getFrameBuffer().clearFB(Color.green);
            Pipeline.render(scene, this.getFrameBuffer());
            repaint();
         }
      };

      app.addKeyListener(app);
      app.addComponentListener(app);
   }
}
